@@ -36,12 +36,13 @@
  *
  */
 
-#include <config.h>
+#include "gear_config.h"
 #include <libgearman/common.h>
 
 #include <libgearman-1.0/visibility.h>
 #include <libgearman/command.h>
-#include <cassert>
+
+#include "libgearman/assert.hpp"
 
 /**
  * Command info. Update GEARMAN_MAX_COMMAND_ARGS to the largest number in the
@@ -49,7 +50,7 @@
  */
 gearman_command_info_st gearmand_command_info_list[GEARMAN_COMMAND_MAX]=
 {
-  { "TEXT",               3, false },
+  { "TEXT",               5, false },
   { "CAN_DO",             1, false },
   { "CANT_DO",            1, false },
   { "RESET_ABILITIES",    0, false },
@@ -89,7 +90,9 @@ gearman_command_info_st gearmand_command_info_list[GEARMAN_COMMAND_MAX]=
   { "GEARMAN_COMMAND_SUBMIT_REDUCE_JOB", 4, true },
   { "GEARMAN_COMMAND_SUBMIT_REDUCE_JOB_BACKGROUND", 4, true },
   { "GEARMAN_COMMAND_GRAB_JOB_ALL",    0, false  },
-  { "GEARMAN_COMMAND_JOB_ASSIGN_ALL",    4, true  }
+  { "GEARMAN_COMMAND_JOB_ASSIGN_ALL",    4, true  },
+  { "GEARMAN_COMMAND_GET_STATUS_UNIQUE", 1, false },
+  { "GEARMAN_COMMAND_STATUS_RES_UNIQUE", 6, false }
 };
 
 gearman_command_info_st *gearman_command_info(gearman_command_t command)
