@@ -2,7 +2,7 @@
  * 
  *  Gearmand client and server library.
  *
- *  Copyright (C) 2011 Data Differential, http://datadifferential.com/
+ *  Copyright (C) 2011-2012 Data Differential, http://datadifferential.com/
  *  Copyright (C) 2008 Brian Aker, Eric Day
  *  All rights reserved.
  *
@@ -41,10 +41,11 @@
  * @brief Connection Definitions
  */
 
-#include <config.h>
+#include "gear_config.h"
 #include <libgearman/common.h>
 
-#include <cassert>
+#include "libgearman/assert.hpp"
+
 #include <cerrno>
 #include <cstdio>
 #include <cstdlib>
@@ -53,102 +54,58 @@
 #include <unistd.h>
 
 #if HAVE_NETINET_TCP_H
-#include <netinet/tcp.h>    /* for TCP_NODELAY */
+# include <netinet/tcp.h>    /* for TCP_NODELAY */
 #endif
 #ifdef HAVE_FCNTL_H
-#include <fcntl.h>
+# include <fcntl.h>
+#endif
+
+#ifndef SOCK_CLOEXEC 
+# define SOCK_CLOEXEC 0
+#endif
+
+#ifndef SOCK_NONBLOCK 
+# define SOCK_NONBLOCK 0
+#endif
+
+#ifndef FD_CLOEXEC
+# define FD_CLOEXEC 0
+#endif
+
+#ifndef MSG_DONTWAIT
+# define MSG_DONTWAIT 0
 #endif
 
 static gearman_return_t gearman_connection_set_option(gearman_connection_st *connection,
                                                       gearman_connection_options_t options,
-                                                      bool value);
-
-
-
-gearman_return_t gearman_connection_st::connect_poll()
+                                                      bool value)
 {
-  struct pollfd fds[1];
-  fds[0].fd= fd;
-  fds[0].events= POLLOUT;
-
-  size_t loop_max= 5;
-
-#if 0
-  if (universal.timeout == 0)
+  switch (options)
   {
-    return gearman_error(universal, GEARMAN_TIMEOUT, "not connected");
-  }
-#endif
+  case GEARMAN_CON_READY:
+    connection->options.ready= value;
+    break;
 
-  while (--loop_max) // Should only loop on cases of ERESTART or EINTR
-  {
-    int error= poll(fds, 1, GEARMAN_DEFAULT_CONNECT_TIMEOUT);
-    switch (error)
-    {
-    case 1:
-      {
-        int err;
-        socklen_t len= sizeof (err);
-        // We replace errno with err if getsockopt() passes, but err has been
-        // set.
-        if (getsockopt(fd, SOL_SOCKET, SO_ERROR, &err, &len) == 0)
-        {
-          // We check the value to see what happened wth the socket.
-          if (err == 0)
-          {
-            return GEARMAN_SUCCESS;
-          }
-          errno= err;
-        }
+  case GEARMAN_CON_PACKET_IN_USE:
+    connection->options.packet_in_use= value;
+    break;
 
-        return gearman_perror(universal, "getsockopt() failed");
-      }
+  case GEARMAN_CON_IGNORE_LOST_CONNECTION:
+    break;
 
-    case 0:
-      {
-        return gearman_error(universal, GEARMAN_TIMEOUT, "timeout occurred while trying to connect");
-      }
+  case GEARMAN_CON_CLOSE_AFTER_FLUSH:
+    break;
 
-    default: // A real error occurred and we need to completely bail
-      switch (get_socket_errno())
-      {
-#ifdef TARGET_OS_LINUX
-      case ERESTART:
-#endif
-      case EINTR:
-        continue;
-
-      case EFAULT:
-      case ENOMEM:
-        return gearman_perror(universal, "poll() failure");
-
-      case EINVAL:
-        return gearman_perror(universal, "RLIMIT_NOFILE exceeded, or if OSX the timeout value was invalid");
-
-      default: // This should not happen
-        if (fds[0].revents & POLLERR)
-        {
-          int err;
-          socklen_t len= sizeof (err);
-          (void)getsockopt(fd, SOL_SOCKET, SO_ERROR, &err, &len);
-          errno= err;
-        }
-        else
-        {
-          errno= get_socket_errno();
-        }
-
-        assert_msg(fd != INVALID_SOCKET, "poll() was passed an invalid file descriptor");
-
-        return gearman_perror(universal, "socket error occurred");
-      }
-    }
+  case GEARMAN_CON_EXTERNAL_FD:
+  case GEARMAN_CON_MAX:
+  default:
+    return GEARMAN_INVALID_COMMAND;
   }
 
-  // This should only be possible from ERESTART or EINTR;
-  return gearman_perror(universal, "connection failed (error should be from either ERESTART or EINTR");
+  return GEARMAN_SUCCESS;
 }
 
+
 /**
  * @addtogroup gearman_connection_static Static Connection Declarations
  * @ingroup gearman_connection
@@ -173,11 +130,15 @@ gearman_connection_st::gearman_connection_st(gearman_universal_st &universal_arg
   recv_buffer_size(0),
   recv_data_size(0),
   recv_data_offset(0),
-  universal(universal_arg)
+  universal(universal_arg),
+  next(NULL),
+  prev(NULL),
+  context(NULL),
+  _addrinfo(NULL),
+  addrinfo_next(NULL),
+  send_buffer_ptr(NULL),
+  _recv_packet(NULL)
 {
-  options.ready= false;
-  options.packet_in_use= false;
-
   if (options_args)
   {
     while (*options_args != GEARMAN_CON_MAX)
@@ -192,15 +153,10 @@ gearman_connection_st::gearman_connection_st(gearman_universal_st &universal_arg
     universal.con_list->prev= this;
   }
   next= universal.con_list;
-  prev= NULL;
   universal.con_list= this;
   universal.con_count++;
 
-  context= NULL;
-  addrinfo= NULL;
-  addrinfo_next= NULL;
   send_buffer_ptr= send_buffer;
-  recv_packet= NULL;
   recv_buffer_ptr= recv_buffer;
   host[0]= 0;
 }
@@ -222,18 +178,16 @@ gearman_connection_st *gearman_connection_create_args(gearman_universal_st& univ
                                                       const char *host, in_port_t port)
 {
   gearman_connection_st *connection= gearman_connection_create(universal, NULL);
-  if (connection == NULL)
+  if (connection)
   {
-    return NULL;
-  }
+    connection->set_host(host, port);
 
-  connection->set_host(host, port);
-
-  if (gearman_failed(connection->lookup()))
-  {
-    gearman_gerror(universal, GEARMAN_GETADDRINFO);
-    delete connection;
-    return NULL;
+    if (gearman_failed(connection->lookup()))
+    {
+      gearman_gerror(universal, GEARMAN_GETADDRINFO);
+      delete connection;
+      return NULL;
+    }
   }
 
   return connection;
@@ -244,18 +198,16 @@ gearman_connection_st *gearman_connection_copy(gearman_universal_st& universal,
 {
   gearman_connection_st *connection= gearman_connection_create(universal, NULL);
 
-  if (connection == NULL)
+  if (connection)
   {
-    return NULL;
+    connection->options.ready= from.options.ready;
+    // @todo Is this right?
+    connection->options.packet_in_use= from.options.packet_in_use;
+
+    strcpy(connection->host, from.host);
+    connection->port= from.port;
   }
 
-  connection->options.ready= from.options.ready;
-  // @todo Is this right?
-  connection->options.packet_in_use= from.options.packet_in_use;
-
-  strcpy(connection->host, from.host);
-  connection->port= from.port;
-
   return connection;
 }
 
@@ -299,40 +251,6 @@ void gearman_connection_st::free_private_packet()
   }
 }
 
-gearman_return_t gearman_connection_set_option(gearman_connection_st *connection,
-                                               gearman_connection_options_t options,
-                                               bool value)
-{
-  switch (options)
-  {
-  case GEARMAN_CON_READY:
-    connection->options.ready= value;
-    break;
-
-  case GEARMAN_CON_PACKET_IN_USE:
-    connection->options.packet_in_use= value;
-    break;
-
-  case GEARMAN_CON_IGNORE_LOST_CONNECTION:
-    break;
-
-  case GEARMAN_CON_CLOSE_AFTER_FLUSH:
-    break;
-
-  case GEARMAN_CON_EXTERNAL_FD:
-  case GEARMAN_CON_MAX:
-  default:
-    return GEARMAN_INVALID_COMMAND;
-  }
-
-  return GEARMAN_SUCCESS;
-}
-
-/**
- * Set socket options for a connection.
- */
-static gearman_return_t _con_setsockopt(gearman_connection_st *connection);
-
 /** @} */
 
 /*
@@ -343,12 +261,30 @@ void gearman_connection_st::set_host(const char *host_arg, const in_port_t port_
 {
   reset_addrinfo();
 
-  strncpy(host, host_arg == NULL ? GEARMAN_DEFAULT_TCP_HOST : host_arg, GEARMAN_NI_MAXHOST);
+  if (host_arg == NULL)
+  {
+    strncpy(host, GEARMAN_DEFAULT_TCP_HOST, GEARMAN_NI_MAXHOST);
+  }
+  else
+  {
+    strncpy(host, host_arg, GEARMAN_NI_MAXHOST);
+  }
+
   host[GEARMAN_NI_MAXHOST - 1]= 0;
 
-  port= in_port_t(port_arg == 0 ? GEARMAN_DEFAULT_TCP_PORT : port_arg);
+  if (port_arg == 0)
+  {
+    port= GEARMAN_DEFAULT_TCP_PORT;
+  }
+  else
+  {
+    port= port_arg;
+  }
 }
 
+/*
+  Do not call error within this function.
+*/
 void gearman_connection_st::close_socket()
 {
   if (fd == INVALID_SOCKET)
@@ -358,20 +294,11 @@ void gearman_connection_st::close_socket()
 
   /* in case of death shutdown to avoid blocking at close_socket() */
   if (shutdown(fd, SHUT_RDWR) == SOCKET_ERROR && get_socket_errno() != ENOTCONN)
-  {
-#if 0
-    gearman_perror(universal, "shutdown");
-    assert(errno != ENOTSOCK);
-#endif
-  }
+  { }
   else
   {
     if (closesocket(fd) == SOCKET_ERROR)
-    {
-#if 0
-      gearman_perror(universal, "close");
-#endif
-    }
+    { }
   }
 
   state= GEARMAN_CON_UNIVERSAL_ADDRINFO;
@@ -386,32 +313,125 @@ void gearman_connection_st::close_socket()
   send_data_offset= 0;
 
   recv_state= GEARMAN_CON_RECV_UNIVERSAL_NONE;
-  if (recv_packet)
-  {
-    gearman_packet_free(recv_packet);
-    recv_packet= NULL;
-  }
+  free_recv_packet();
 
   recv_buffer_ptr= recv_buffer;
   recv_buffer_size= 0;
+
+  options.server_options_sent= false;
+
+  // created_id_next is incremented for every outbound packet (except status).
+  // created_id is incremented for every response packet received, and also when
+  // no packets are received due to an error. There are lots of such error paths
+  // and it seems simpler to just reset these both to zero when a connection is
+  // 'closed'.
+  created_id= 0;
+  created_id_next= 0;
+}
+
+void gearman_connection_st::free_recv_packet()
+{
+  if (_recv_packet)
+  {
+    gearman_packet_free(recv_packet());
+    _recv_packet= NULL;
+  }
 }
 
 void gearman_connection_st::reset_addrinfo()
 {
-  if (addrinfo)
+  if (_addrinfo)
   {
-    freeaddrinfo(addrinfo);
-    addrinfo= NULL;
+    freeaddrinfo(_addrinfo);
+    _addrinfo= NULL;
   }
 
   addrinfo_next= NULL;
 }
 
+
+/*
+ * The send_packet() method does not only send the passed-in packet_arg. If there are any server options
+ * established, and they haven't yet been sent over, then these options are sent over first.
+ * Only if that succeeds is the packet_arg sent.
+ * The reason for this is server options are only set once by the client/worker. In the older code, this
+ * resulted in them being sent over exactly once. If the connection was dropped and rebuilt, then the options
+ * were not sent over again, rendering them moot. This way, we're guaranteed that the options are always sent
+ * at least once to a connected server.
+ */
 gearman_return_t gearman_connection_st::send_packet(const gearman_packet_st& packet_arg, const bool flush_buffer)
+{
+  if (options.server_options_sent == false)
+  {
+    for (gearman_server_options_st* head= universal.server_options_list;
+         head;
+         head= head->next)
+    {
+      gearman_packet_st message;
+      const void *args[]= { head->_option };
+      size_t args_size[]= { head->_option_length };
+      gearman_return_t ret= gearman_packet_create_args(universal, message, GEARMAN_MAGIC_REQUEST,
+                                                       GEARMAN_COMMAND_OPTION_REQ, args, args_size, 1);
+
+      if (gearman_failed(ret))
+      {
+        gearman_packet_free(&message);
+        gearman_error(universal, GEARMAN_MEMORY_ALLOCATION_FAILURE, "gearman_packet_create_args()");
+        return GEARMAN_MEMORY_ALLOCATION_FAILURE;
+      }
+
+      PUSH_BLOCKING(universal);
+      OptionCheck check(universal);
+      ret= _send_packet(message, true);
+      if (gearman_failed(ret))
+      {
+        gearman_packet_free(&message);
+        gearman_error(universal, ret, "Failed to send server-options packet");
+        return ret;
+      }
+
+      options.packet_in_use= true;
+      gearman_packet_st *packet_ptr= receiving(_packet, ret, true);
+      if (packet_ptr == NULL)
+      {
+        gearman_packet_free(&message);
+        options.packet_in_use= false;
+        gearman_error(universal, ret, "Failed in receiving()");
+        return ret;
+      }
+
+      if (gearman_failed(ret) ||
+          gearman_failed(ret= check.success(this)))
+      {
+        gearman_packet_free(&message);
+        free_private_packet();
+        reset_recv_packet();
+        gearman_error(universal, ret, "receiving()");
+        return ret;
+      }
+
+      free_private_packet();
+      reset_recv_packet();
+      gearman_packet_free(&message);
+    }
+
+    options.server_options_sent= true;
+  }
+
+  return _send_packet(packet_arg, flush_buffer);
+}
+
+/*
+ * This is the real implementation that actually sends a packet. Read the comments for send_packet() for why
+ * that is. Note that this is a private method. External callers should only call send_packet().
+ */
+gearman_return_t gearman_connection_st::_send_packet(const gearman_packet_st& packet_arg, const bool flush_buffer)
 {
   switch (send_state)
   {
   case GEARMAN_CON_SEND_STATE_NONE:
+    assert_msg(packet_arg.universal, "send_packet() was to execute against a packet with no universal");
+    universal_reset_error(*(packet_arg.universal));
     if (packet_arg.options.complete == false)
     {
       return gearman_error(universal, GEARMAN_INVALID_PACKET, "packet not complete");
@@ -420,8 +440,8 @@ gearman_return_t gearman_connection_st::send_packet(const gearman_packet_st& pac
     /* Pack first part of packet, which is everything but the payload. */
     while (1)
     {
-      gearman_return_t rc;
       { // Scoping to shut compiler up about switch/case jump
+        gearman_return_t rc;
         size_t send_size= gearman_packet_pack(packet_arg,
                                               send_buffer +send_buffer_size,
                                               GEARMAN_SEND_BUFFER_SIZE -send_buffer_size, rc);
@@ -502,7 +522,7 @@ gearman_return_t gearman_connection_st::send_packet(const gearman_packet_st& pac
     send_data_size= packet_arg.data_size;
 
     /* If this is NULL, then gearman_connection_send_data function will be used. */
-    if (not packet_arg.data)
+    if (packet_arg.data == NULL)
     {
       send_state= GEARMAN_CON_SEND_UNIVERSAL_FLUSH_DATA;
       return GEARMAN_SUCCESS;
@@ -513,14 +533,14 @@ gearman_return_t gearman_connection_st::send_packet(const gearman_packet_st& pac
     if (send_buffer_size < GEARMAN_SEND_BUFFER_SIZE)
     {
       memcpy(send_buffer,
-             static_cast<char *>(const_cast<void *>(packet_arg.data)) + send_data_offset,
+             (const char*)(packet_arg.data) +send_data_offset,
              send_buffer_size);
       send_data_size= 0;
       send_data_offset= 0;
       break;
     }
 
-    send_buffer_ptr= static_cast<char *>(const_cast<void *>(packet_arg.data)) + send_data_offset;
+    send_buffer_ptr= (const char*)(size_t(packet_arg.data) +send_data_offset);
     send_state= GEARMAN_CON_SEND_UNIVERSAL_FLUSH_DATA;
 
   case GEARMAN_CON_SEND_UNIVERSAL_FLUSH:
@@ -551,7 +571,7 @@ size_t gearman_connection_st::send_and_flush(const void *data, size_t data_size,
     return gearman_error(universal, GEARMAN_DATA_TOO_LARGE, "data too large");
   }
 
-  send_buffer_ptr= static_cast<char *>(const_cast<void *>(data));
+  send_buffer_ptr= (const char*)data;
   send_buffer_size= data_size;
 
   *ret_ptr= flush();
@@ -561,27 +581,30 @@ size_t gearman_connection_st::send_and_flush(const void *data, size_t data_size,
 
 gearman_return_t gearman_connection_st::lookup()
 {
-  if (addrinfo)
-  {
-    freeaddrinfo(addrinfo);
-    addrinfo= NULL;
-  }
+  reset_addrinfo();
 
-  char port_str[GEARMAN_NI_MAXSERV];
-  snprintf(port_str, GEARMAN_NI_MAXSERV, "%hu", uint16_t(port));
+  char port_str[GEARMAN_NI_MAXSERV]= { 0 };
+  int port_str_length;
+  if (size_t(port_str_length= snprintf(port_str, sizeof(port_str), "%hu", uint16_t(port))) >= sizeof(port_str))
+  {
+    return gearman_universal_set_error(universal, GEARMAN_MEMORY_ALLOCATION_FAILURE, GEARMAN_AT, "snprintf(%d)", port_str_length);
+  }
 
   struct addrinfo ai;
   memset(&ai, 0, sizeof(struct addrinfo));
   ai.ai_socktype= SOCK_STREAM;
   ai.ai_protocol= IPPROTO_TCP;
 
+  assert_msg(_addrinfo == NULL, "Programmer error, reset_addrinfo() is either broke, or was not called.");
   int ret;
-  if ((ret= getaddrinfo(host, port_str, &ai, &(addrinfo))))
+  if ((ret= getaddrinfo(host, port_str, &ai, &(_addrinfo))))
   {
+    reset_addrinfo();
     return gearman_universal_set_error(universal, GEARMAN_GETADDRINFO, GEARMAN_AT, "getaddrinfo:%s", gai_strerror(ret));
   }
 
-  addrinfo_next= addrinfo;
+  addrinfo_next= _addrinfo;
+  assert(addrinfo_next);
   state= GEARMAN_CON_UNIVERSAL_CONNECT;
 
   return GEARMAN_SUCCESS;
@@ -615,7 +638,22 @@ gearman_return_t gearman_connection_st::flush()
         return gearman_universal_set_error(universal, GEARMAN_COULD_NOT_CONNECT, GEARMAN_AT, "%s:%hu", host, uint16_t(port));
       }
 
-      fd= socket(addrinfo_next->ai_family, addrinfo_next->ai_socktype, addrinfo_next->ai_protocol);
+      // rewrite tye if HAVE_SOCK_CLOEXEC
+      {
+        int type= addrinfo_next->ai_socktype;
+        if (SOCK_CLOEXEC)
+        {
+          type|= SOCK_CLOEXEC;
+        }
+
+        if (SOCK_NONBLOCK)
+        {
+          type|= SOCK_NONBLOCK; 
+        }
+
+        fd= socket(addrinfo_next->ai_family, type, addrinfo_next->ai_protocol);
+      }
+
       if (fd == INVALID_SOCKET)
       {
         state= GEARMAN_CON_UNIVERSAL_ADDRINFO;
@@ -623,7 +661,7 @@ gearman_return_t gearman_connection_st::flush()
       }
 
       {
-        gearman_return_t gret= _con_setsockopt(this);
+        gearman_return_t gret= set_socket_options();
         if (gearman_failed(gret))
         {
           close_socket();
@@ -640,35 +678,32 @@ gearman_return_t gearman_connection_st::flush()
           break;
         }
 
-        if (errno == EAGAIN || errno == EINTR)
+        switch (errno)
         {
-          continue;
-        }
+          // Treat as an async connect
+        case EINTR:
+        case EINPROGRESS:
+            state= GEARMAN_CON_UNIVERSAL_CONNECTING;
+            break;
 
-        if (errno == EINPROGRESS)
-        {
-          gearman_return_t gret= connect_poll();
-          if (gearman_failed(gret))
-          {
-            assert_msg(universal.error.rc != GEARMAN_SUCCESS, "Programmer error, connect_poll() returned an error, but it was not set");
-            close_socket();
-            return gret;
-          }
-
-          state= GEARMAN_CON_UNIVERSAL_CONNECTING;
-          break;
-        }
-
-        if (errno == ECONNREFUSED || errno == ENETUNREACH || errno == ETIMEDOUT)
-        {
-          state= GEARMAN_CON_UNIVERSAL_CONNECT;
+        case ECONNREFUSED:
+        case ENETUNREACH:
+        case ETIMEDOUT:
           addrinfo_next= addrinfo_next->ai_next;
+
+          // We will treat this as an error but retry the address
+        case EAGAIN:
+          state= GEARMAN_CON_UNIVERSAL_CONNECT;
+          close_socket();
           break;
+
+        default:
+          gearman_perror(universal, "connect");
+          close_socket();
+          return GEARMAN_COULD_NOT_CONNECT;
         }
 
-        gearman_perror(universal, "connect");
-        close_socket();
-        return GEARMAN_COULD_NOT_CONNECT;
+        break;
       }
 
       if (state != GEARMAN_CON_UNIVERSAL_CONNECTING)
@@ -714,40 +749,44 @@ gearman_return_t gearman_connection_st::flush()
     case GEARMAN_CON_UNIVERSAL_CONNECTED:
       while (send_buffer_size != 0)
       {
-        ssize_t write_size= ::send(fd, send_buffer_ptr, send_buffer_size, 
-                                   gearman_universal_is_non_blocking(universal) ? MSG_NOSIGNAL| MSG_DONTWAIT : MSG_NOSIGNAL);
+        ssize_t write_size= ::send(fd, send_buffer_ptr, send_buffer_size, MSG_NOSIGNAL);
 
         if (write_size == 0) // Zero value on send()
         { }
         else if (write_size == -1)
         {
-          if (errno == EAGAIN)
+          switch (errno)
           {
-            set_events(POLLOUT);
-
-            if (gearman_universal_is_non_blocking(universal))
+#if defined(EWOULDBLOCK) && EWOULDBLOCK != EAGAIN
+          case EWOULDBLOCK:
+#endif
+          case EAGAIN:
             {
-              gearman_gerror(universal, GEARMAN_IO_WAIT);
-              return GEARMAN_IO_WAIT;
-            }
+              set_events(POLLOUT);
 
-            gearman_return_t gret= gearman_wait(universal);
-            if (gearman_failed(gret))
+              if (gearman_universal_is_non_blocking(universal))
+              {
+                return gearman_gerror(universal, GEARMAN_IO_WAIT);
+              }
+
+              gearman_return_t gret= gearman_wait(universal);
+              if (gearman_failed(gret))
+              {
+                return gret;
+              }
+
+              continue;
+            }
+          case EPIPE:
+          case ECONNRESET:
+          case EHOSTDOWN:
             {
-              return gret;
+              gearman_return_t ret= gearman_perror(universal, "lost connection to server during send");
+              close_socket();
+              return ret;
             }
-
-            continue;
-          }
-          else if (errno == EINTR)
-          {
-            continue;
-          }
-          else if (errno == EPIPE || errno == ECONNRESET || errno == EHOSTDOWN)
-          {
-            gearman_return_t ret= gearman_perror(universal, "lost connection to server during send");
-            close_socket();
-            return ret;
+          default:
+            break;
           }
 
           gearman_return_t ret= gearman_perror(universal, "send");
@@ -797,15 +836,14 @@ gearman_packet_st *gearman_connection_st::receiving(gearman_packet_st& packet_ar
   case GEARMAN_CON_RECV_UNIVERSAL_NONE:
     if (state != GEARMAN_CON_UNIVERSAL_CONNECTED)
     {
-      gearman_error(universal, GEARMAN_NOT_CONNECTED, "not connected");
-      ret= GEARMAN_NOT_CONNECTED;
+      ret= gearman_error(universal, GEARMAN_NOT_CONNECTED, "not connected");
       return NULL;
     }
 
-    recv_packet= gearman_packet_create(universal, &packet_arg);
-    if (recv_packet == NULL)
+    _recv_packet= gearman_packet_create(universal, &packet_arg);
+    if (_recv_packet == NULL)
     {
-      ret= GEARMAN_MEMORY_ALLOCATION_FAILURE;
+      ret= gearman_error(universal, GEARMAN_MEMORY_ALLOCATION_FAILURE, "gearman_packet_create()");
       return NULL;
     }
 
@@ -814,9 +852,10 @@ gearman_packet_st *gearman_connection_st::receiving(gearman_packet_st& packet_ar
   case GEARMAN_CON_RECV_UNIVERSAL_READ:
     while (1)
     {
+      // If we have data, see if it is a complete packet
       if (recv_buffer_size > 0)
       {
-        size_t recv_size= gearman_packet_unpack(*recv_packet,
+        size_t recv_size= gearman_packet_unpack(*(recv_packet()),
                                                 recv_buffer_ptr,
                                                 recv_buffer_size, ret);
         recv_buffer_ptr+= recv_size;
@@ -840,7 +879,7 @@ gearman_packet_st *gearman_connection_st::receiving(gearman_packet_st& packet_ar
       }
       recv_buffer_ptr= recv_buffer;
 
-      size_t recv_size= recv_socket(recv_buffer + recv_buffer_size, GEARMAN_RECV_BUFFER_SIZE - recv_buffer_size, ret);
+      size_t recv_size= recv_socket(recv_buffer +recv_buffer_size, GEARMAN_RECV_BUFFER_SIZE -recv_buffer_size, ret);
       if (gearman_failed(ret))
       {
         return NULL;
@@ -857,7 +896,7 @@ gearman_packet_st *gearman_connection_st::receiving(gearman_packet_st& packet_ar
 
     recv_data_size= packet_arg.data_size;
 
-    if (not recv_data)
+    if (recv_data == false )
     {
       recv_state= GEARMAN_CON_RECV_STATE_READ_DATA;
       break;
@@ -865,9 +904,9 @@ gearman_packet_st *gearman_connection_st::receiving(gearman_packet_st& packet_ar
 
     assert(packet_arg.universal);
     packet_arg.data= gearman_malloc((*packet_arg.universal), packet_arg.data_size);
-    if (not packet_arg.data)
+    if (packet_arg.data == NULL)
     {
-      ret= GEARMAN_MEMORY_ALLOCATION_FAILURE;
+      ret= gearman_error(universal, GEARMAN_MEMORY_ALLOCATION_FAILURE, "gearman_malloc((*packet_arg.universal), packet_arg.data_size)");
       close_socket();
       return NULL;
     }
@@ -878,7 +917,7 @@ gearman_packet_st *gearman_connection_st::receiving(gearman_packet_st& packet_ar
   case GEARMAN_CON_RECV_STATE_READ_DATA:
     while (recv_data_size)
     {
-      (void)receiving(static_cast<uint8_t *>(const_cast<void *>(packet_arg.data)) +
+      (void)receive_data(static_cast<uint8_t *>(const_cast<void *>(packet_arg.data)) +
                       recv_data_offset,
                       packet_arg.data_size -recv_data_offset, ret);
       if (gearman_failed(ret))
@@ -891,13 +930,13 @@ gearman_packet_st *gearman_connection_st::receiving(gearman_packet_st& packet_ar
     break;
   }
 
-  gearman_packet_st *tmp_packet_arg= recv_packet;
-  recv_packet= NULL;
+  gearman_packet_st *tmp_packet_arg= recv_packet();
+  reset_recv_packet();
 
   return tmp_packet_arg;
 }
 
-size_t gearman_connection_st::receiving(void *data, size_t data_size, gearman_return_t& ret)
+size_t gearman_connection_st::receive_data(void *data, size_t data_size, gearman_return_t& ret)
 {
   size_t recv_size= 0;
 
@@ -908,14 +947,20 @@ size_t gearman_connection_st::receiving(void *data, size_t data_size, gearman_re
   }
 
   if ((recv_data_size - recv_data_offset) < data_size)
+  {
     data_size= recv_data_size - recv_data_offset;
+  }
 
   if (recv_buffer_size > 0)
   {
     if (recv_buffer_size < data_size)
+    {
       recv_size= recv_buffer_size;
+    }
     else
+    {
       recv_size= data_size;
+    }
 
     memcpy(data, recv_buffer_ptr, recv_size);
     recv_buffer_ptr+= recv_size;
@@ -949,12 +994,12 @@ size_t gearman_connection_st::recv_socket(void *data, size_t data_size, gearman_
 
   while (1)
   {
-    read_size= ::recv(fd, data, data_size, 0);
+    read_size= ::recv(fd, data, data_size, MSG_NOSIGNAL);
+
     if (read_size == 0)
     {
-      gearman_error(universal, GEARMAN_LOST_CONNECTION, "lost connection to server (EOF)");
+      ret= gearman_error(universal, GEARMAN_LOST_CONNECTION, "lost connection to server (EOF)");
       close_socket();
-      ret= GEARMAN_LOST_CONNECTION;
 
       return 0;
     }
@@ -966,30 +1011,19 @@ size_t gearman_connection_st::recv_socket(void *data, size_t data_size, gearman_
 
         if (gearman_universal_is_non_blocking(universal))
         {
-          gearman_gerror(universal, GEARMAN_IO_WAIT);
-          ret= GEARMAN_IO_WAIT;
+          ret= gearman_gerror(universal, GEARMAN_IO_WAIT);
           return 0;
         }
 
         ret= gearman_wait(universal);
 
-        if (ret == GEARMAN_SHUTDOWN_GRACEFUL)
-        {
-          ret= gearman_kill(gearman_universal_id(universal), GEARMAN_INTERRUPT);
-
-          if (gearman_failed(ret))
-          {
-            ret= GEARMAN_SHUTDOWN;
-          }
-        }
-        else if (ret == GEARMAN_SHUTDOWN)
-        {
-          close_socket();
-          return 0;
-        }
-
         if (gearman_failed(ret))
         {
+          if (ret == GEARMAN_SHUTDOWN)
+          {
+            close_socket();
+          }
+
           return 0;
         }
 
@@ -1001,13 +1035,11 @@ size_t gearman_connection_st::recv_socket(void *data, size_t data_size, gearman_
       }
       else if (errno == EPIPE || errno == ECONNRESET || errno == EHOSTDOWN)
       {
-        gearman_perror(universal, "lost connection to server during read");
-        ret= GEARMAN_LOST_CONNECTION;
+        ret= gearman_perror(universal, "lost connection to server during read");
       }
       else
       {
-        gearman_perror(universal, "read");
-        ret= GEARMAN_ERRNO;
+        ret= gearman_perror(universal, "recv");
       }
 
       close_socket();
@@ -1024,7 +1056,9 @@ size_t gearman_connection_st::recv_socket(void *data, size_t data_size, gearman_
 void gearman_connection_st::set_events(short arg)
 {
   if ((events | arg) == events)
+  {
     return;
+  }
 
   events|= arg;
 }
@@ -1032,7 +1066,9 @@ void gearman_connection_st::set_events(short arg)
 void gearman_connection_st::set_revents(short arg)
 {
   if (arg)
+  {
     options.ready= true;
+  }
 
   revents= arg;
   events&= short(~arg);
@@ -1042,101 +1078,148 @@ void gearman_connection_st::set_revents(short arg)
  * Static Definitions
  */
 
-static gearman_return_t _con_setsockopt(gearman_connection_st *connection)
+gearman_return_t gearman_connection_st::set_socket_options()
 {
-  int ret;
-  struct linger linger;
-  struct timeval waittime;
-
-  ret= 1;
-  ret= setsockopt(connection->fd, IPPROTO_TCP, TCP_NODELAY, &ret,
-                  socklen_t(sizeof(int)));
-  if (ret == -1 && errno != EOPNOTSUPP)
+  if (SOCK_CLOEXEC == 0)
   {
-    gearman_perror(connection->universal, "setsockopt(TCP_NODELAY)");
-    return GEARMAN_ERRNO;
-  }
-
-  linger.l_onoff= 1;
-  linger.l_linger= GEARMAN_DEFAULT_SOCKET_TIMEOUT;
-  ret= setsockopt(connection->fd, SOL_SOCKET, SO_LINGER, &linger,
-                  socklen_t(sizeof(struct linger)));
-  if (ret == -1)
-  {
-    gearman_perror(connection->universal, "setsockopt(SO_LINGER)");
-    return GEARMAN_ERRNO;
-  }
-
-  waittime.tv_sec= GEARMAN_DEFAULT_SOCKET_TIMEOUT;
-  waittime.tv_usec= 0;
-  ret= setsockopt(connection->fd, SOL_SOCKET, SO_SNDTIMEO, &waittime,
-                  socklen_t(sizeof(struct timeval)));
-  if (ret == -1 && errno != ENOPROTOOPT)
-  {
-    gearman_perror(connection->universal, "setsockopt(SO_SNDTIMEO)");
-    return GEARMAN_ERRNO;
-  }
-
-  {
-    int optval= 1;
-    ret= setsockopt(connection->fd, SOL_SOCKET, SO_KEEPALIVE, &optval, sizeof(optval));
-    if (ret == -1 && errno != ENOPROTOOPT)
+    if (FD_CLOEXEC)
     {
-      gearman_perror(connection->universal, "setsockopt(SO_RCVTIMEO)");
+      int flags;
+      do 
+      {
+        flags= fcntl(fd, F_GETFD, 0);
+      } while (flags == -1 and (errno == EINTR or errno == EAGAIN));
+
+      if (flags != -1)
+      {
+        int rval;
+        do
+        { 
+          rval= fcntl (fd, F_SETFD, flags | FD_CLOEXEC);
+        } while (rval == -1 && (errno == EINTR or errno == EAGAIN));
+        // we currently ignore the case where rval is -1
+      }
+    }
+  }
+
+  {
+    int ret= 1;
+    ret= setsockopt(fd, IPPROTO_TCP, TCP_NODELAY, &ret,
+                    socklen_t(sizeof(int)));
+    if (ret == -1 && errno != EOPNOTSUPP)
+    {
+      gearman_perror(universal, "setsockopt(TCP_NODELAY)");
       return GEARMAN_ERRNO;
     }
   }
 
-
-  ret= setsockopt(connection->fd, SOL_SOCKET, SO_RCVTIMEO, &waittime,
-                  socklen_t(sizeof(struct timeval)));
-  if (ret == -1 && errno != ENOPROTOOPT)
   {
-    gearman_perror(connection->universal, "setsockopt(SO_RCVTIMEO)");
-    return GEARMAN_ERRNO;
+    struct linger linger;
+    linger.l_onoff= 1;
+    linger.l_linger= GEARMAN_DEFAULT_SOCKET_TIMEOUT;
+    int ret= setsockopt(fd, SOL_SOCKET, SO_LINGER, &linger,
+                        socklen_t(sizeof(struct linger)));
+    if (ret == -1)
+    {
+      gearman_perror(universal, "setsockopt(SO_LINGER)");
+      return GEARMAN_ERRNO;
+    }
   }
 
-  ret= GEARMAN_DEFAULT_SOCKET_SEND_SIZE;
-  ret= setsockopt(connection->fd, SOL_SOCKET, SO_SNDBUF, &ret, socklen_t(sizeof(int)));
-  if (ret == -1)
+  if (0)
   {
-    gearman_perror(connection->universal, "setsockopt(SO_SNDBUF)");
-    return GEARMAN_ERRNO;
+    struct timeval waittime;
+    waittime.tv_sec= GEARMAN_DEFAULT_SOCKET_TIMEOUT;
+    waittime.tv_usec= 0;
+    int ret= setsockopt(fd, SOL_SOCKET, SO_SNDTIMEO, &waittime,
+                        socklen_t(sizeof(struct timeval)));
+    if (ret == -1 && errno != ENOPROTOOPT)
+    {
+      gearman_perror(universal, "setsockopt(SO_SNDTIMEO)");
+      return GEARMAN_ERRNO;
+    }
+
+    ret= setsockopt(fd, SOL_SOCKET, SO_RCVTIMEO, &waittime,
+                    socklen_t(sizeof(struct timeval)));
+    if (ret == -1 && errno != ENOPROTOOPT)
+    {
+      gearman_perror(universal, "setsockopt(SO_RCVTIMEO)");
+      return GEARMAN_ERRNO;
+    }
   }
 
-#if defined(__MACH__) && defined(__APPLE__) || defined(__FreeBSD__)
   {
-    ret= 1;
-    setsockopt(connection->fd, SOL_SOCKET, SO_NOSIGPIPE, static_cast<void *>(&ret), sizeof(int));
+    int optval= 1;
+    int ret= setsockopt(fd, SOL_SOCKET, SO_KEEPALIVE, &optval, sizeof(optval));
+    if (ret == -1 && errno != ENOPROTOOPT)
+    {
+      gearman_perror(universal, "setsockopt(SO_KEEPALIVE)");
+      return GEARMAN_ERRNO;
+    }
+  }
+
+  {
+    int ret= GEARMAN_DEFAULT_SOCKET_SEND_SIZE;
+    ret= setsockopt(fd, SOL_SOCKET, SO_SNDBUF, &ret, socklen_t(sizeof(int)));
+    if (ret == -1)
+    {
+      gearman_perror(universal, "setsockopt(SO_SNDBUF)");
+      return GEARMAN_ERRNO;
+    }
+  }
+
+#if defined(SO_NOSIGPIPE)
+  if (SO_NOSIGPIPE)
+  {
+    int ret= 1;
+    ret= setsockopt(fd, SOL_SOCKET, SO_NOSIGPIPE, static_cast<void *>(&ret), sizeof(int));
 
     // This is not considered a fatal error 
     if (ret == -1)
     {
-      gearman_perror(connection->universal, "setsockopt(SO_NOSIGPIPE)");
+      gearman_perror(universal, "setsockopt(SO_NOSIGPIPE)");
     }
   }
 #endif
 
-  ret= GEARMAN_DEFAULT_SOCKET_RECV_SIZE;
-  ret= setsockopt(connection->fd, SOL_SOCKET, SO_RCVBUF, &ret, socklen_t(sizeof(int)));
-  if (ret == -1)
   {
-    gearman_perror(connection->universal, "setsockopt(SO_RCVBUF)");
-    return GEARMAN_ERRNO;
+    int ret= GEARMAN_DEFAULT_SOCKET_RECV_SIZE;
+    ret= setsockopt(fd, SOL_SOCKET, SO_RCVBUF, &ret, socklen_t(sizeof(int)));
+    if (ret == -1)
+    {
+      gearman_perror(universal, "setsockopt(SO_RCVBUF)");
+      return GEARMAN_ERRNO;
+    }
   }
 
-  ret= fcntl(connection->fd, F_GETFL, 0);
-  if (ret == -1)
+  // If SOCK_NONBLOCK doesn't work, just enable non_block via fcntl
+  if (SOCK_NONBLOCK == 0)
   {
-    gearman_perror(connection->universal, "fcntl(F_GETFL)");
-    return GEARMAN_ERRNO;
-  }
+    int flags;
+    do
+    {
+      flags= fcntl(fd, F_GETFL, 0);
+    } while (flags == -1 and (errno == EINTR or errno == EAGAIN));
 
-  ret= fcntl(connection->fd, F_SETFL, ret | O_NONBLOCK);
-  if (ret == -1)
-  {
-    gearman_perror(connection->universal, "fcntl(F_SETFL)");
-    return GEARMAN_ERRNO;
+    if (flags == -1)
+    {
+      gearman_perror(universal, "fcntl(F_GETFL)");
+      return GEARMAN_ERRNO;
+    }
+    else if ((flags & O_NONBLOCK) == 0)
+    {
+      int retval;
+      do
+      {
+        retval= fcntl(fd, F_SETFL, flags | O_NONBLOCK);
+      } while (retval == -1 and (errno == EINTR or errno == EAGAIN));
+
+      if (retval == -1)
+      {
+        gearman_perror(universal, "fcntl(F_SETFL)");
+        return GEARMAN_ERRNO;
+      }
+    }
   }
 
   return GEARMAN_SUCCESS;
