@@ -2,8 +2,8 @@
  * 
  *  Gearmand client and server library.
  *
- *  Copyright (C) 2011 Data Differential, http://datadifferential.com/
- *  Copyright (C) 2008 Brian Aker, Eric Day
+ *  Copyright (C) 2011-2012 Data Differential, http://datadifferential.com/
+ *  Copyright (C) 2008 Brian Aker, Eric Day 
  *  All rights reserved.
  *
  *  Redistribution and use in source and binary forms, with or without
@@ -42,10 +42,12 @@
  * @brief Gearman State Definitions
  */
 
-#include <config.h>
+#include "gear_config.h"
 #include <libgearman/common.h>
 
-#include <cassert>
+#include "libgearman/assert.hpp"
+#include "libgearman/server_options.hpp"
+
 #include <cerrno>
 #include <cstdarg>
 #include <cstdio>
@@ -53,13 +55,12 @@
 #include <cstring>
 #include <cctype>
 #include <unistd.h>
+#include <memory>
 
 void gearman_universal_initialize(gearman_universal_st &self, const gearman_universal_options_t *options)
 {
   { // Set defaults on all options.
-    self.options.dont_track_packets= false;
     self.options.non_blocking= false;
-    self.options.stored_non_blocking= false;
   }
 
   if (options)
@@ -82,6 +83,7 @@ void gearman_universal_initialize(gearman_universal_st &self, const gearman_univ
   self.timeout= -1;
   self.con_list= NULL;
   self.packet_list= NULL;
+  self.server_options_list= NULL;
   self.pfds= NULL;
   self.log_fn= NULL;
   self.log_context= NULL;
@@ -133,20 +135,22 @@ void gearman_universal_clone(gearman_universal_st &destination, const gearman_un
   }
 
   (void)gearman_universal_set_option(destination, GEARMAN_NON_BLOCKING, source.options.non_blocking);
-  (void)gearman_universal_set_option(destination, GEARMAN_DONT_TRACK_PACKETS, source.options.dont_track_packets);
 
   destination.timeout= source.timeout;
 
   destination._namespace= gearman_string_clone(source._namespace);
+  destination.verbose= source.verbose;
+  destination.log_fn= source.log_fn;
+  destination.log_context= source.log_context;
 
   for (gearman_connection_st *con= source.con_list; con; con= con->next)
   {
     if (gearman_connection_copy(destination, *con) == NULL)
     {
-      gearman_universal_free(destination);
       return;
     }
   }
+  assert(destination.con_count == source.con_count);
 }
 
 void gearman_universal_free(gearman_universal_st &universal)
@@ -159,6 +163,13 @@ void gearman_universal_free(gearman_universal_st &universal)
   {
     // created realloc()
     free(universal.pfds);
+    universal.pfds= NULL;
+  }
+
+  // clean-up server options
+  while (universal.server_options_list)
+  {
+    delete universal.server_options_list;
   }
 }
 
@@ -171,7 +182,6 @@ gearman_return_t gearman_universal_set_option(gearman_universal_st &self, gearma
     break;
 
   case GEARMAN_DONT_TRACK_PACKETS:
-    self.options.dont_track_packets= value;
     break;
 
   case GEARMAN_MAX:
@@ -232,7 +242,12 @@ void gearman_reset(gearman_universal_st& universal)
   }
 }
 
-gearman_return_t gearman_flush_all(gearman_universal_st& universal)
+/*
+ * Flush all shouldn't return any error, because there's no way to indicate
+ * which connection experienced an issue. Error detection is better done in gearman_wait()
+ * after flushing all the connections here.
+ */
+void gearman_flush_all(gearman_universal_st& universal)
 {
   for (gearman_connection_st *con= universal.con_list; con; con= con->next)
   {
@@ -241,14 +256,8 @@ gearman_return_t gearman_flush_all(gearman_universal_st& universal)
       continue;
     }
 
-    gearman_return_t ret= con->flush();
-    if (gearman_failed(ret) and ret != GEARMAN_IO_WAIT)
-    {
-      return ret;
-    }
+    con->flush();
   }
-
-  return GEARMAN_SUCCESS;
 }
 
 gearman_return_t gearman_wait(gearman_universal_st& universal)
@@ -268,7 +277,7 @@ gearman_return_t gearman_wait(gearman_universal_st& universal)
     }
 
     universal.pfds= pfds;
-    universal.pfds_size= con_count;
+    universal.pfds_size= int(con_count);
   }
   else
   {
@@ -279,7 +288,9 @@ gearman_return_t gearman_wait(gearman_universal_st& universal)
   for (gearman_connection_st *con= universal.con_list; con; con= con->next)
   {
     if (con->events == 0)
+    {
       continue;
+    }
 
     pfds[x].fd= con->fd;
     pfds[x].events= con->events;
@@ -314,6 +325,9 @@ gearman_return_t gearman_wait(gearman_universal_st& universal)
       case EINTR:
         continue;
 
+      case EINVAL:
+        return gearman_perror(universal, "RLIMIT_NOFILE exceeded, or if OSX the timeout value was invalid");
+
       default:
         return gearman_perror(universal, "poll");
       }
@@ -324,8 +338,9 @@ gearman_return_t gearman_wait(gearman_universal_st& universal)
 
   if (ret == 0)
   {
-    gearman_error(universal, GEARMAN_TIMEOUT, "timeout reached, no servers were available");
-    return GEARMAN_TIMEOUT;
+    return gearman_universal_set_error(universal, GEARMAN_TIMEOUT, GEARMAN_AT,
+                                       "timeout reached, %u servers were poll(), no servers were available, pipe:%s",
+                                       uint32_t(x), have_shutdown_pipe ? "true" : "false");
   }
 
   x= 0;
@@ -336,11 +351,14 @@ gearman_return_t gearman_wait(gearman_universal_st& universal)
       continue;
     }
 
-    int err;
-    socklen_t len= sizeof (err);
-    if (getsockopt(con->fd, SOL_SOCKET, SO_ERROR, &err, &len) == 0)
+    if (pfds[x].revents & (POLLERR | POLLHUP | POLLNVAL))
     {
-      con->cached_errno= err;
+      int err;
+      socklen_t len= sizeof (err);
+      if (getsockopt(pfds[x].fd, SOL_SOCKET, SO_ERROR, &err, &len) == 0)
+      {
+        con->cached_errno= err;
+      }
     }
 
     con->set_revents(pfds[x].revents);
@@ -354,14 +372,23 @@ gearman_return_t gearman_wait(gearman_universal_st& universal)
     ssize_t read_length= read(universal.wakeup_fd[0], buffer, sizeof(buffer));
     if (read_length > 0)
     {
+      gearman_return_t local_ret= gearman_kill(gearman_universal_id(universal), GEARMAN_INTERRUPT);
+      if (gearman_failed(local_ret))
+      {
+        return GEARMAN_SHUTDOWN;
+      }
+
       return GEARMAN_SHUTDOWN_GRACEFUL;
     }
+
     if (read_length == 0)
     {
       return GEARMAN_SHUTDOWN;
     }
 
+#if 0
     perror("shudown read");
+#endif
     // @todo figure out what happens in an error
   }
 
@@ -386,26 +413,9 @@ gearman_connection_st *gearman_ready(gearman_universal_st& universal)
   return NULL;
 }
 
-/**
-  @note _push_blocking is only used for echo (and should be fixed
-  when tricky flip/flop in IO is fixed).
-*/
-static inline void _push_blocking(gearman_universal_st& universal, bool &orig_block_universal)
-{
-  orig_block_universal= universal.options.non_blocking;
-  universal.options.non_blocking= false;
-}
-
-static inline void _pop_non_blocking(gearman_universal_st& universal, bool orig_block_universal)
-{
-  universal.options.non_blocking= orig_block_universal;
-}
-
 gearman_return_t gearman_set_identifier(gearman_universal_st& universal,
                                         const char *id, size_t id_size)
 {
-  bool orig_block_universal;
-
   if (id == NULL)
   {
     return gearman_error(universal, GEARMAN_INVALID_ARGUMENT, "id was NULL");
@@ -433,74 +443,127 @@ gearman_return_t gearman_set_identifier(gearman_universal_st& universal,
   gearman_return_t ret= gearman_packet_create_args(universal, packet, GEARMAN_MAGIC_REQUEST,
                                                    GEARMAN_COMMAND_SET_CLIENT_ID,
                                                    (const void**)&id, &id_size, 1);
-  if (gearman_failed(ret))
+  if (gearman_success(ret))
   {
-#if 0
-    assert_msg(universal.error.rc != GEARMAN_SUCCESS, "Programmer error, error returned but not recorded");
-#endif
-    gearman_packet_free(&packet);
-    return ret;
+    PUSH_BLOCKING(universal);
+
+    for (gearman_connection_st *con= universal.con_list; con; con= con->next)
+    {
+      gearman_return_t local_ret= con->send_packet(packet, true);
+      if (gearman_failed(local_ret))
+      {
+        ret= local_ret;
+      }
+
+    }
   }
 
-  _push_blocking(universal, orig_block_universal);
+  gearman_packet_free(&packet);
+
+  return ret;
+}
+
+EchoCheck::EchoCheck(gearman_universal_st& universal_,
+    const void *workload_, const size_t workload_size_) :
+    _universal(universal_),
+    _workload(workload_),
+    _workload_size(workload_size_)
+{
+}
+
+gearman_return_t EchoCheck::success(gearman_connection_st* con)
+{
+  if (con->_packet.command != GEARMAN_COMMAND_ECHO_RES)
+  {
+    return gearman_error(_universal, GEARMAN_INVALID_COMMAND, "Wrong command sent in response to ECHO request");
+  }
+
+  if (con->_packet.data_size != _workload_size or
+      memcmp(_workload, con->_packet.data, _workload_size))
+  {
+    return gearman_error(_universal, GEARMAN_ECHO_DATA_CORRUPTION, "corruption during echo");
+  }
+
+  return GEARMAN_SUCCESS;
+}
+
+OptionCheck::OptionCheck(gearman_universal_st& universal_):
+    _universal(universal_)
+{
+}
+
+gearman_return_t OptionCheck::success(gearman_connection_st* con)
+{
+  if (con->_packet.command == GEARMAN_COMMAND_ERROR)
+  {
+    return gearman_error(_universal, GEARMAN_INVALID_SERVER_OPTION, "invalid server option");
+  }
+
+  return GEARMAN_SUCCESS;
+}
+
+static gearman_return_t connection_loop(gearman_universal_st& universal,
+                                        const gearman_packet_st& message,
+                                        Check& check)
+{
+  gearman_return_t ret= GEARMAN_SUCCESS;
 
   for (gearman_connection_st *con= universal.con_list; con; con= con->next)
   {
-    ret= con->send_packet(packet, true);
+    ret= con->send_packet(message, true);
     if (gearman_failed(ret))
     {
 #if 0
       assert_msg(con->universal.error.rc != GEARMAN_SUCCESS, "Programmer error, error returned but not recorded");
 #endif
-      goto exit;
+      break;
     }
 
     con->options.packet_in_use= true;
     gearman_packet_st *packet_ptr= con->receiving(con->_packet, ret, true);
+    if (packet_ptr == NULL)
+    {
+      assert(&con->_packet == universal.packet_list);
+      con->options.packet_in_use= false;
+      break;
+    }
+
+    assert(packet_ptr == &con->_packet);
     if (gearman_failed(ret))
     {
 #if 0
       assert_msg(con->universal.error.rc != GEARMAN_SUCCESS, "Programmer error, error returned but not recorded");
 #endif
       con->free_private_packet();
-      con->recv_packet= NULL;
+      con->reset_recv_packet();
 
-      goto exit;
+      break;
     }
     assert(packet_ptr);
 
-    if (con->_packet.data_size != id_size or
-        memcmp(id, con->_packet.data, id_size))
+    if (gearman_failed(ret= check.success(con)))
     {
 #if 0
       assert_msg(con->universal.error.rc != GEARMAN_SUCCESS, "Programmer error, error returned but not recorded");
 #endif
       con->free_private_packet();
-      con->recv_packet= NULL;
-      ret= gearman_error(universal, GEARMAN_ECHO_DATA_CORRUPTION, "corruption during echo");
+      con->reset_recv_packet();
 
-      goto exit;
+      break;
     }
 
-    con->recv_packet= NULL;
+    con->reset_recv_packet();
     con->free_private_packet();
   }
 
-  ret= GEARMAN_SUCCESS;
-
-exit:
-  gearman_packet_free(&packet);
-  _pop_non_blocking(universal, orig_block_universal);
-
   return ret;
 }
 
+
 gearman_return_t gearman_echo(gearman_universal_st& universal,
                               const void *workload,
                               size_t workload_size)
 {
-  bool orig_block_universal;
-
   if (workload == NULL)
   {
     return gearman_error(universal, GEARMAN_INVALID_ARGUMENT, "workload was NULL");
@@ -516,135 +579,27 @@ gearman_return_t gearman_echo(gearman_universal_st& universal,
     return gearman_error(universal, GEARMAN_ARGUMENT_TOO_LARGE,  "workload_size was greater then GEARMAN_MAX_ECHO_SIZE");
   }
 
-  gearman_packet_st packet;
-  gearman_return_t ret= gearman_packet_create_args(universal, packet, GEARMAN_MAGIC_REQUEST,
+  gearman_packet_st message;
+  gearman_return_t ret= gearman_packet_create_args(universal, message, GEARMAN_MAGIC_REQUEST,
                                                    GEARMAN_COMMAND_ECHO_REQ,
                                                    &workload, &workload_size, 1);
-  if (gearman_failed(ret))
+  if (gearman_success(ret))
   {
-#if 0
-    assert_msg(universal.error.rc != GEARMAN_SUCCESS, "Programmer error, error returned but not recorded");
-#endif
-    gearman_packet_free(&packet);
-    return ret;
+    PUSH_BLOCKING(universal);
+
+    EchoCheck check(universal, workload, workload_size);
+    ret= connection_loop(universal, message, check);
   }
-
-  _push_blocking(universal, orig_block_universal);
-
-  for (gearman_connection_st *con= universal.con_list; con; con= con->next)
+  else
   {
-    ret= con->send_packet(packet, true);
-    if (gearman_failed(ret))
-    {
-#if 0
-      assert_msg(con->universal.error.rc != GEARMAN_SUCCESS, "Programmer error, error returned but not recorded");
-#endif
-      goto exit;
-    }
-
-    con->options.packet_in_use= true;
-    gearman_packet_st *packet_ptr= con->receiving(con->_packet, ret, true);
-    if (gearman_failed(ret))
-    {
-#if 0
-      assert_msg(con->universal.error.rc != GEARMAN_SUCCESS, "Programmer error, error returned but not recorded");
-#endif
-      con->free_private_packet();
-      con->recv_packet= NULL;
-
-      goto exit;
-    }
-    assert(packet_ptr);
-
-    if (con->_packet.data_size != workload_size or
-        memcmp(workload, con->_packet.data, workload_size))
-    {
-#if 0
-      assert_msg(con->universal.error.rc != GEARMAN_SUCCESS, "Programmer error, error returned but not recorded");
-#endif
-      con->free_private_packet();
-      con->recv_packet= NULL;
-      ret= gearman_error(universal, GEARMAN_ECHO_DATA_CORRUPTION, "corruption during echo");
-
-      goto exit;
-    }
-
-    con->recv_packet= NULL;
-    con->free_private_packet();
-  }
-
-  ret= GEARMAN_SUCCESS;
-
-exit:
-  gearman_packet_free(&packet);
-  _pop_non_blocking(universal, orig_block_universal);
-
-  return ret;
-}
-
-bool gearman_request_option(gearman_universal_st &universal,
-                            gearman_string_t &option)
-{
-  bool orig_block_universal;
-
-  const void *args[]= { gearman_c_str(option) };
-  size_t args_size[]= { gearman_size(option) };
-
-  gearman_packet_st packet;
-  gearman_return_t ret= gearman_packet_create_args(universal, packet, GEARMAN_MAGIC_REQUEST,
-                                                   GEARMAN_COMMAND_OPTION_REQ,
-                                                   args, args_size, 1);
-  if (gearman_failed(ret))
-  {
-    gearman_packet_free(&packet);
+    gearman_packet_free(&message);
     gearman_error(universal, GEARMAN_MEMORY_ALLOCATION_FAILURE, "gearman_packet_create_args()");
     return ret;
   }
 
-  _push_blocking(universal, orig_block_universal);
+  gearman_packet_free(&message);
 
-  for (gearman_connection_st *con= universal.con_list; con != NULL; con= con->next)
-  {
-    ret= con->send_packet(packet, true);
-    if (gearman_failed(ret))
-    {
-      goto exit;
-    }
-
-    gearman_packet_st recv_packet;
-    assert(con->recv_state == GEARMAN_CON_RECV_UNIVERSAL_NONE);
-    gearman_packet_st *packet_ptr= con->receiving(recv_packet, ret, true);
-    if (ret == GEARMAN_NOT_CONNECTED)
-    {
-      goto exit;
-    }
-    else if (gearman_failed(ret))
-    {
-      con->recv_packet= NULL;
-      gearman_packet_free(&recv_packet);
-      goto exit;
-    }
-    assert(packet_ptr);
-
-    if (packet_ptr->command == GEARMAN_COMMAND_ERROR)
-    {
-      con->recv_packet= NULL;
-      gearman_packet_free(&recv_packet);
-      ret= gearman_error(universal, GEARMAN_INVALID_ARGUMENT, "invalid server option");
-
-      goto exit;
-    }
-
-    gearman_packet_free(&recv_packet);
-  }
-
-  ret= GEARMAN_SUCCESS;
-
-exit:
-  gearman_packet_free(&packet);
-  _pop_non_blocking(universal, orig_block_universal);
-
-  return gearman_success(ret);
+  return ret;
 }
 
 void gearman_free_all_packets(gearman_universal_st &universal)
@@ -669,5 +624,17 @@ gearman_id_t gearman_universal_id(gearman_universal_st &universal)
 void gearman_universal_set_namespace(gearman_universal_st& universal, const char *namespace_key, size_t namespace_key_size)
 {
   gearman_string_free(universal._namespace);
-  universal._namespace= gearman_string_create(NULL, namespace_key, namespace_key_size);
+  if (namespace_key)
+  {
+    universal._namespace= gearman_string_create(NULL, namespace_key, namespace_key_size);
+  }
+  else
+  {
+    universal._namespace= NULL;
+  }
+}
+
+const char *gearman_univeral_namespace(gearman_universal_st& universal)
+{
+  return gearman_string_value(universal._namespace);
 }
